
using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Xml;
using System.Linq;
using System.Collections.Generic;
using DataContracts;
using System.Text.RegularExpressions;

namespace Ato.EN.IntegrationServices.CodeGenerationTFND
{

    public partial class TFND2019Validator
    {
        /// <summary>
        /// The max parameter name length is restricted by the SBR1 and SBR2 schemas - the lowest common denominator of SBR 1 is used as the default
        /// </summary>
        private int _maxParameterNameLength;

        /// <summary>
        /// The max parameter value length is restricted by the SBR1 and SBR2 schemas
        /// </summary>
        private int _maxParameterValueLength;

        /// <summary>
        /// The SBR1 and ebms3 schemas do not allow parameter names or values to be the empty string
        /// </summary>
        private string _emptyParameterValue;

        /// <summary>
        /// Initializes a new instance of the <see cref="TFND2019Validator" /> class.
        /// </summary>
        /// <param name="maxParameterNameLength">Maximum length of the parameter name.</param>
        /// <param name="maxParameterValueLength">Maximum length of the parameter value.</param>
        /// <param name="emptyParameterValue">This value will be used in place of any parameter values that result in a null or empty value.</param>
        public TFND2019Validator(int maxParameterNameLength = 20, int maxParameterValueLength = 4096, string emptyParameterValue = "EMPTY")
        {
            _maxParameterNameLength = maxParameterNameLength;
            _maxParameterValueLength = maxParameterValueLength;
            _emptyParameterValue = emptyParameterValue;
        }

        #region Functions
        private static IEnumerable<string> Union(IEnumerable<string> list1, IEnumerable<string> list2)
        {
            IEnumerable<string> response;

            if (list1 == null && list2 == null)
            {
                response = null;
            }
            else if (list1 == null)
            {
                response = list2.Distinct();
            }
            else if (list2 == null)
            {
                response = list1.Distinct();
            }
            else
            {
                response = list1.Union(list2);
            }

            return response;
        }

        private static bool IsMatch(int? field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(Convert.ToString(field.Value), expression, options);
        }

        private static bool IsMatch(string field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(field, expression, options);
        }

        // This is just context and tuple counts where they are integer values - easier that changing the parsing logic to just return the value
        private static int Count(int count)
        {
            return count;
        }


        private static int Count<T>(IEnumerable<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static int Count<T>(ICollection<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static bool exists(bool value)
        {
            return value;
        }

        private static string GetValueOrEmpty(bool? val)
        {
            return (val.HasValue) ? val.ToString().ToLower() : string.Empty;
        }

        private static string GetValueOrEmptyDate(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-dd") : string.Empty;
        }

        private static string GetValueOrEmptyDate(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-dd");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-ddTHH:mm:ssZ");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-ddTHH:mm:ssZ") : string.Empty;
        }

        private static string GetValueOrEmpty(string val)
        {
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmpty(decimal? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        private static string GetValueOrEmpty(int? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

         /// <summary>
         /// Get string value between [first] a and [last] b.
         /// </summary>
        public static string Between(string value, string a, string b)
        {
            int posA = value.IndexOf(a);
            int posB = value.LastIndexOf(b);
            if (posA == -1)
            {
                return "";
            }
            if (posB == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= posB)
            {
                return "";
            }
            return value.Substring(adjustedPosA, posB - adjustedPosA);
        }

         /// <summary>
         /// Get string value after [first] a.
         /// </summary>
        public static string Before(string value, string a)
        {
            int posA = value.IndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            return value.Substring(0, posA);
        }

         /// <summary>
         /// Get string value after [last] a.
         /// </summary>
        public static string After(string value, string a)
        {
            int posA = value.LastIndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= value.Length)
            {
                return "";
            }
            return value.Substring(adjustedPosA);
        }

        private static int Length(object field)
        {
            if (field == null)
            return 0;
            else
            return field.ToString().Trim().Length;
        }

        private static bool NotSameValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(params object[] nodes)
        {
            if (nodes == null)
                return false;

            return (from x in nodes select x).Distinct().Count() != 1;
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(params object[] nodes)
        {
           if (nodes == null)
                return false;

            nodes = nodes.Where(x => x != null).ToArray();
            return !((from x in nodes select x).Distinct().Count() == nodes.Count());
        
        }

        private int DuplicateValueIndex(IEnumerable<object> values)
        {
            int response = 0;
            var hashset = new HashSet<object>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private int DuplicateValueIndex<T>(IEnumerable<T?> values) where T : struct
        {
            int response = 0;
            var hashset = new HashSet<T?>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private static bool IsDate(object value)
        {
            DateTime dateValue;
            return (value != null && DateTime.TryParse(value.ToString(), out dateValue));
        }

        public DateTime AsDate(string dateAsString)
        {
            DateTime response = DateTime.MinValue;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date;
            }

            return response;
        }

        public DateTime Date(DateTime? datetime)
        {                   
            return datetime.GetValueOrDefault().Date;
        }

				//The logic in After function expects "---" for day and "--" for month. 
				//Since hyphen was missing the date always returned null
        public DateTime? ConvertToDate(int day, int month, int year)
        {
					return ConvertToDate(day == 0 ? null : "---" + day.ToString(), month == 0 ? null : "--" + month.ToString(), year == 0 ? null : year.ToString());
        }

        public DateTime? ConvertToDate(string day, string month, string year)
        {
            DateTime? response;
            DateTime result;

            if (year == null || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        public DateTime? ConvertToDate(string day, string month, int year)
        {
            DateTime? response;
            DateTime result;

            if (year == 0 || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year.ToString() + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        private static int Day(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Day;
            }

            return response;
        }


        private static int Day(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Day;

        }


        private static string Month(string dateAsString)
        {
            string response = null;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.ToString("MMMM");
            }
            else
            {
                return "NotAMonth";
            }

            return response;
        }

        private static string Month(DateTime? date)
        {
            if (date == null)
                return "NotAMonth";
            else
                return date.Value.ToString("MMMM");
        }


        private static int MonthAsInt(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Month;
            }

            return response;
        }

        private static int MonthAsInt(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            return date.Value.Month;
        }


        private static int Year(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Year;
            }

            return response;
        }

        private static int Year(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Year;
        }


        private static int CurrentFinancialYear()
        {
            return DateToFinancialYear(DateTime.Now, 7);
        }

        private static int FinancialYear(string dateAsString)
        {
            return DateToFinancialYear(dateAsString, 7);
        }

        private static int FinancialYear(DateTime? date)
        {
            return DateToFinancialYear(date, 7);
        }

        private static int DateToFinancialYear(string dateAsString, int startingMonth)
        {
            int response = 0;
            DateTime date;
            if (DateTime.TryParse(dateAsString, out date))
            {
                response = DateToFinancialYear(date, startingMonth);
            }

            return response;
        }
        private static int DateToFinancialYear(DateTime? date, int startingMonth)
        {
            int response;
            if (date == null)
            {
                response = 0;
            }
            else
            {
                int year = date.Value.Year;
                int month = date.Value.Month;

                if (startingMonth > month)
                    response = year;
                else
                    response = year + 1;
            }
            return response;
        }


        private static int FBTYear(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
                response = FBTYear(date);
            return response;
        }

        private static int FBTYear(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            else
            {
                if (date.Value.Month > 3)
                    return date.Value.Year + 1;
                else
                    return date.Value.Year;
            }
        }

        private static DateTime? AddMonthsToDate(DateTime? dateTime, int months)
        {
            return dateTime == null ? null : (DateTime?)dateTime.Value.AddMonths(months);
        }

        private static bool IsNumeric(object value)
        {
            decimal numbervalue;
            return (value != null && decimal.TryParse(value.ToString(), out numbervalue));
        }

        private static bool NotMonetary(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(int? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }

        private static bool NotNumeric(long? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }


        private static bool OutsideRange(decimal field, decimal expression, int range)
        {
            bool response;

            response = (field < (expression - range)) || (field > (expression + range));

            return response;
        }


        private static bool FailsUSIAlgorithm(string usi, string abn)
        {
            bool response = true;
            if (usi == null || abn == null)
            {
                response = false;
            }
            else
            {
                usi = usi.Trim();
                abn = abn.Trim();
                if (usi.Length == 14)
                {
                    int numeric;
                    if (usi.Substring(0, 11) == abn && int.TryParse(usi.Substring(11, 3), out numeric))
                        response = false;

                }
                else if (usi.Length == 9)
                {

                    if (Regex.IsMatch(usi, @"^([a-zA-Z]{3}\d{4}[a-zA-Z]{2})$"))
                        response = false;
                }
            }
            return response;
        }

        private static bool FailsTANAlgorithm(string tan)
        {
            bool response;
            decimal decimalTan;

            if (tan == null)
                return false;

            tan = tan.Trim();

            if (!decimal.TryParse(tan, out decimalTan))
                return true;

            if (tan.Length != 8)
                return true;

            decimal tanSum =
                7 * int.Parse(tan.Substring(0, 1)) +
                9 * int.Parse(tan.Substring(1, 1)) +
                8 * int.Parse(tan.Substring(2, 1)) +
                4 * int.Parse(tan.Substring(3, 1)) +
                6 * int.Parse(tan.Substring(4, 1)) +
                3 * int.Parse(tan.Substring(5, 1)) +
                5 * int.Parse(tan.Substring(6, 1)) +
                1 * int.Parse(tan.Substring(7, 1));

            if ((tanSum % 11) == 0)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsABNAlgorithm(string abn)
        {
            bool response;
            decimal decimalAbn;

            if (abn == null)
                return false;

            abn = abn.Trim();

            if (!decimal.TryParse(abn, out decimalAbn))
                return true;

            if (abn.Length != 11)
                return true;

            decimal abnSum =
                10 * (int.Parse(abn.Substring(0, 1)) - 1) +
                1 * int.Parse(abn.Substring(1, 1)) +
                3 * int.Parse(abn.Substring(2, 1)) +
                5 * int.Parse(abn.Substring(3, 1)) +
                7 * int.Parse(abn.Substring(4, 1)) +
                9 * int.Parse(abn.Substring(5, 1)) +
                11 * int.Parse(abn.Substring(6, 1)) +
                13 * int.Parse(abn.Substring(7, 1)) +
                15 * int.Parse(abn.Substring(8, 1)) +
                17 * int.Parse(abn.Substring(9, 1)) +
                19 * int.Parse(abn.Substring(10, 1));

            if ((abnSum % 89) == 0)
                response = false;
            else
                response = true;

            return response;
        }
        private static bool FailsACNAlgorithm(string acn)
        {
            bool response;
            decimal decimalAbn;
            if (acn == null)
                return false;

            acn = acn.Trim();

            if (!decimal.TryParse(acn, out decimalAbn))
                return true;

            if (acn.Length != 9)
                return true;

            decimal abnSum =
                8 * int.Parse(acn.Substring(0, 1)) +
                7 * int.Parse(acn.Substring(1, 1)) +
                6 * int.Parse(acn.Substring(2, 1)) +
                5 * int.Parse(acn.Substring(3, 1)) +
                4 * int.Parse(acn.Substring(4, 1)) +
                3 * int.Parse(acn.Substring(5, 1)) +
                2 * int.Parse(acn.Substring(6, 1)) +
                1 * int.Parse(acn.Substring(7, 1));

            decimal checkDigit = int.Parse(acn.Substring(8, 1));
            decimal acnRemainder = abnSum % 10;

            if (((10 - acnRemainder) % 10) == checkDigit)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsTFNAlgorithm(string tfn)
        {
            bool response;
            decimal decimalTfn;

            if (tfn == null)
                return false;

            tfn = tfn.Trim();
            tfn = Regex.Replace(tfn, "^0+", "");

            if (!decimal.TryParse(tfn, out decimalTfn))
                return true;

            if (tfn.Length < 8)
                return true;


            decimal tfn1To7Sum =
                1 * int.Parse(tfn.Substring(0, 1)) +
                4 * int.Parse(tfn.Substring(1, 1)) +
                3 * int.Parse(tfn.Substring(2, 1)) +
                7 * int.Parse(tfn.Substring(3, 1)) +
                5 * int.Parse(tfn.Substring(4, 1)) +
                8 * int.Parse(tfn.Substring(5, 1)) +
                6 * int.Parse(tfn.Substring(6, 1));

            decimal tfn8 = 9 * int.Parse(tfn.Substring(7, 1));

            if (tfn.Length == 8)
            {
                decimal tFNLg8WSum9 = 10 * int.Parse(tfn.Substring(7, 1));
                decimal tFNLg8WSum = tfn1To7Sum + tFNLg8WSum9;

                if ((tFNLg8WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else if (tfn.Length == 9)
            {
                decimal tfn9 = 10 * int.Parse(tfn.Substring(8, 1));
                decimal tFNLg9WSum = tfn1To7Sum + tfn8 + tfn9;

                if ((tFNLg9WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else
            {
                response = true;
            }

            return response;
        }


        private static decimal ConditionalValue(bool expression, decimal? trueVal, decimal? falseVal)
        {
             return expression ? trueVal.GetValueOrDefault() : falseVal.GetValueOrDefault();
        }

        private static decimal AsNumeric(string value)
        {
             decimal numberValue;
             decimal.TryParse(value, out numberValue);
             return numberValue;
        }

        private static bool RegexMatch(int? field, string expression, string flags="")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static bool RegexMatch(string field, string expression, string flags="")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static RegexOptions GetRegexOption(string flags)
        {
            RegexOptions options = RegexOptions.None;

            char[] characters = flags.ToCharArray();

            foreach (char character in characters)
            {
                switch (character)
                {
                    case 'i':                        
                        options = options | RegexOptions.IgnoreCase;                       
                        break;
                    case 'm':                        
                        options = options | RegexOptions.Multiline;                       
                        break;
                    case 's':                        
                            options = options | RegexOptions.Singleline;                       
                        break;
                    case 'n':                        
                            options = options | RegexOptions.ExplicitCapture;                       
                        break;
                    case 'x':                       
                            options = options | RegexOptions.IgnorePatternWhitespace;                       
                        break;
                }
            }

            return options;
        }

        private static string Substring(object field, int start, int length)
        {
            return field.ToString().Substring(start, length);
        }
        
        /// <summary>
        /// Returns an occurrence index as [occurrenceIndex] of occurrenceIndex > 0, otherwise the empty string
        /// </summary>
        /// <param name="occurrenceIndex">Index of the occurrence.</param>
        /// <returns>Occurrence in XPath [#] format</returns>
        public string OccurrenceIndex(int occurrenceIndex)
        {
            return occurrenceIndex > 0 ? "[" + occurrenceIndex + "]" : "";
        }

        #endregion // Functions

        public TFND2019 ConsumedReport { get; private set; }

        private static Dictionary<string, ProcessMessageDocument> _processMessageDocuments = new Dictionary<string,ProcessMessageDocument>();

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public List<ProcessMessageDocument> ValidateReport(TFND2019 report)
        {

            List<ProcessMessageDocument> response = new List<ProcessMessageDocument>();
            ProcessMessageDocument processMessage;
            ProcessMessageParameter parameter;
            bool assertion;

            this.ConsumedReport = report;

            // ------------------------------------------------------------------------------
            // Validations are now done in code.
            // This version supports full validation of a report - for those rules that have successfully parsed and could be generated.
            // Generated Validations include:
            //   - Validation logic in C#.
            //   - Production of errors in EBMS, SBR and SWS format.
            // 
            // The generation at this stage does not support the following - and has to be completed manually (later versions of the generation will do this for you)
            //   - Business Rules that did not parse.  The ESR team will help support you where this happens as an effort is being made to rectify these
            //   - You will get TODO tasks for those that the parser could not cope with
            // ------------------------------------------------------------------------------

    
            #region VR.ATO.TFND.000113
    
            /*  VR.ATO.TFND.000113
            When the Payer address is within Australia (Country Code is ‘au’ or not supplied), the State / Territory must be provided.
    
            Legacy Rule Format:
            ((^TFND79 = NULL OR ContainsSet(^TFND79,'"au"')) AND (^TFND32 = NULL))

            Technical Business Rule Format:
            ((^TFND79 = NULL OR ContainsSet(^TFND79,'"au"')) AND (^TFND32 = NULL))
    
            Data Elements:
    
            ^TFND32 = TFND:RP:AddressDetailsBusiness:AddressDetails.StateOrTerritory.Code
    
            ^TFND79 = TFND:RP:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = ((report.TFND79 == null || IsMatch(report.TFND79, @"au",RegexOptions.IgnoreCase)) && report.TFND32 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000112", Severity = ProcessMessageSeverity.Error,
                    Description = @"For an Australian address, the state or territory must be provided",
                    Location = "/tns:TFND/tns:RP/tns:AddressDetailsBusiness/tns:StateOrTerritoryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000113"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND79", Value = GetValueOrEmpty(report.TFND79) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND32", Value = GetValueOrEmpty(report.TFND32) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000113
    
            #region VR.ATO.TFND.000153
    
            /*  VR.ATO.TFND.000153
            The Supplier Australian Business Number (ABN) has failed the ABN algorithm check.
    
            Legacy Rule Format:
            FailsABNAlgorithm(^TFND117)

            Technical Business Rule Format:
            FailsABNAlgorithm(^TFND117)
    
            Data Elements:
    
            ^TFND117 = TFND:INT:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = FailsABNAlgorithm(report.TFND117);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000153", Severity = ProcessMessageSeverity.Error,
                    Description = @"Supplier ABN is invalid.",
                    Location = "/tns:TFND/tns:INT/tns:AustralianBusinessNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000153"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND117", Value = GetValueOrEmpty(report.TFND117) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000153
    
            #region VR.ATO.TFND.000156
    
            /*  VR.ATO.TFND.000156
            The Supplier Tax Agent Number has failed the TAN algorithm check.
    
            Legacy Rule Format:
            FailsTANAlgorithm(^TFND116)

            Technical Business Rule Format:
            FailsTANAlgorithm(^TFND116)
    
            Data Elements:
    
            ^TFND116 = TFND:INT:Identifiers.TaxAgentNumber.Identifier
            */
            assertion = FailsTANAlgorithm(report.TFND116);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000156", Severity = ProcessMessageSeverity.Error,
                    Description = @"Supplier Tax Agent Number is invalid.",
                    Location = "/tns:TFND/tns:INT/tns:TaxAgentNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000156"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND116", Value = GetValueOrEmpty(report.TFND116) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000156
    
            #region VR.ATO.TFND.000159
    
            /*  VR.ATO.TFND.000159
            Supplier business address country code must consist of a valid Country Code
    
            Legacy Rule Format:
            (^TFND75 <> NULL AND (NotInDomain(^TFND75, 'COUNTRY CODES')))

            Technical Business Rule Format:
            (^TFND75 <> NULL AND (NotInDomain(^TFND75, 'COUNTRY CODES')))
    
            Data Elements:
    
            ^TFND75 = TFND:INT:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TFND75 != null && !(IsMatch(report.TFND75, @"^(ad|ae|af|ag|ai|al|am|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bl|bm|bn|bo|bq|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cu|cv|cw|cx|cy|cz|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mf|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|ss|st|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tr|tt|tv|tw|tz|ua|ug|um|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|za|zm|zw)$",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000159", Severity = ProcessMessageSeverity.Error,
                    Description = @"Supplier business address country code is invalid.",
                    Location = "/tns:TFND/tns:INT/tns:AddressDetailsBusiness/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000159"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND75", Value = GetValueOrEmpty(report.TFND75) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000159
    
            #region VR.ATO.TFND.000165
    
            /*  VR.ATO.TFND.000165
            When the Supplier address is overseas, the State / Territory must not be present.
    
            Legacy Rule Format:
            ((^TFND75 <> NULL AND NotContainsSet(^TFND75,'"au"')) AND (^TFND19 <> NULL))

            Technical Business Rule Format:
            ((^TFND75 <> NULL AND NotContainsSet(^TFND75,'"au"')) AND (^TFND19 <> NULL))
    
            Data Elements:
    
            ^TFND19 = TFND:INT:AddressDetailsBusiness:AddressDetails.StateOrTerritory.Code
    
            ^TFND75 = TFND:INT:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TFND75 != null && !(IsMatch(report.TFND75, @"au",RegexOptions.IgnoreCase)) && report.TFND19 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500011", Severity = ProcessMessageSeverity.Error,
                    Description = @"State and postcode must not be present for non-Australian addresses",
                    LongDescription = @"Country name is present with State or postcode in the address field.  Determine whether the address is Australian or overseas and correct accordingly.",
                    Location = "/tns:TFND/tns:INT/tns:AddressDetailsBusiness/tns:StateOrTerritoryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000165"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND75", Value = GetValueOrEmpty(report.TFND75) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND19", Value = GetValueOrEmpty(report.TFND19) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000165
    
            #region VR.ATO.TFND.000168
    
            /*  VR.ATO.TFND.000168
            When the Supplier address is overseas, the Postcode must not be present.
    
            Legacy Rule Format:
            ((^TFND75 <> NULL AND NotContainsSet(^TFND75,'"au"')) AND (^TFND14 <> NULL))

            Technical Business Rule Format:
            ((^TFND75 <> NULL AND NotContainsSet(^TFND75,'"au"')) AND (^TFND14 <> NULL))
    
            Data Elements:
    
            ^TFND14 = TFND:INT:AddressDetailsBusiness:AddressDetails.Postcode.Text
    
            ^TFND75 = TFND:INT:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TFND75 != null && !(IsMatch(report.TFND75, @"au",RegexOptions.IgnoreCase)) && report.TFND14 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000480", Severity = ProcessMessageSeverity.Error,
                    Description = @"Country code must be blank or set to Australia",
                    Location = "/tns:TFND/tns:INT/tns:AddressDetailsBusiness/tns:PostcodeT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000168"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND75", Value = GetValueOrEmpty(report.TFND75) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND14", Value = GetValueOrEmpty(report.TFND14) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000168
    
            #region VR.ATO.TFND.000171
    
            /*  VR.ATO.TFND.000171
            When Supplier Business Address Details is provided, and the address is within Australia (Country Code is ‘au’ or not supplied), the Postcode must be provided.
    
            Legacy Rule Format:
            ^TFND206 <> NULL AND (ContainsSet(^TFND75,'"au"') OR ^TFND75 = NULL) AND ^TFND14 = NULL

            Technical Business Rule Format:
            ^TFND206 <> NULL AND (ContainsSet(^TFND75,'"au"') OR ^TFND75 = NULL) AND ^TFND14 = NULL
    
            Data Elements:
    
            ^TFND14 = TFND:INT:AddressDetailsBusiness:AddressDetails.Postcode.Text
    
            ^TFND75 = TFND:INT:AddressDetailsBusiness:AddressDetails.Country.Code
    
            ^TFND206 = TFND:INT:AddressDetailsBusiness
            */
            assertion = (report.InT_AddressDetailsBusinessCollectionExists != false && (IsMatch(report.TFND75, @"au",RegexOptions.IgnoreCase) || report.TFND75 == null) && report.TFND14 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000009", Severity = ProcessMessageSeverity.Error,
                    Description = @"Postcode is required for an Australian address",
                    Location = "/tns:TFND/tns:INT/tns:AddressDetailsBusiness/tns:PostcodeT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000171"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND75", Value = GetValueOrEmpty(report.TFND75) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND14", Value = GetValueOrEmpty(report.TFND14) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000171
    
            #region VR.ATO.TFND.000177
    
            /*  VR.ATO.TFND.000177
            If provided, must be a valid format.
    
            Legacy Rule Format:
            (^TFND22 <> BLANK AND NOT IsValidEmail(^TFND22))

            Technical Business Rule Format:
            (^TFND22 <> BLANK AND NOT IsValidEmail(^TFND22))
    
            Data Elements:
    
            ^TFND22 = TFND:INT:ElectronicContact:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TFND22) != true &&  !(IsMatch(report.TFND22, @"^\S.*@.+\.\S+$") ));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500029", Severity = ProcessMessageSeverity.Error,
                    Description = @"Email Address contains invalid text",
                    LongDescription = @"A text character must be one of the following: A to Z a to z 0 to 9 ! @ $ % & * ( ) - _ = [ ] ; : ' "" , . ? / or a space character.",
                    Location = "/tns:TFND/tns:INT/tns:ElectronicContact/tns:ElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000177"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND22", Value = GetValueOrEmpty(report.TFND22) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000177
    
            #region VR.ATO.TFND.000180
    
            /*  VR.ATO.TFND.000180
            The Supplier date of declaration signed cannot be later than today.
    
            Legacy Rule Format:
            ^TFND131 > Today()

            Technical Business Rule Format:
            ^TFND131 > Today()
    
            Data Elements:
    
            ^TFND131 = TFND:INT:Declaration:Declaration.Signature.Date
            */
            assertion = (report.TFND131.GetValueOrDefault() > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000180", Severity = ProcessMessageSeverity.Error,
                    Description = @"The date entered for Supplier date of declaration signed must not be later than today's date.",
                    Location = "/tns:TFND/tns:INT/tns:Declaration/tns:SignatureD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000180"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND131", Value = GetValueOrEmptyDate(report.TFND131) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000180
    
            #region VR.ATO.TFND.000183
    
            /*  VR.ATO.TFND.000183
            Supplier declaration accepted indicator must be 'TRUE'.
    
            Legacy Rule Format:
            ((^TFND260 <> NULL) AND (^TFND130 <> TRUE))

            Technical Business Rule Format:
            ((^TFND260 <> NULL) AND (^TFND130 <> TRUE))
    
            Data Elements:
    
            ^TFND130 = TFND:INT:Declaration:Declaration.StatementAccepted.Indicator
    
            ^TFND260 = TFND:INT:Declaration
            */
            assertion = (report.InT_DeclarationCollectionExists != false && report.TFND130 != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000183", Severity = ProcessMessageSeverity.Error,
                    Description = @"Supplier declaration accepted indicator must be 'TRUE'.",
                    Location = "/tns:TFND/tns:INT/tns:Declaration/tns:StatementAcceptedI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000183"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND130", Value = GetValueOrEmpty(report.TFND130) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000183
    
            #region VR.ATO.TFND.000186
    
            /*  VR.ATO.TFND.000186
            Either the Payer Australian Business Number or the Payer Withholding Payer Number must be provided.
    
            Legacy Rule Format:
            (^TFND118 = NULL AND ^TFND261 = NULL) OR (^TFND118 <> NULL AND ^TFND261 <> NULL)

            Technical Business Rule Format:
            (^TFND118 = NULL AND ^TFND261 = NULL) OR (^TFND118 <> NULL AND ^TFND261 <> NULL)
    
            Data Elements:
    
            ^TFND118 = TFND:RP:Identifiers.AustralianBusinessNumber.Identifier
    
            ^TFND261 = TFND:RP:Identifiers.WithholdingPayerNumber.Identifier
            */
            assertion = (report.TFND118 == null && report.TFND261 == null || report.TFND118 != null && report.TFND261 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200010", Severity = ProcessMessageSeverity.Error,
                    Description = @"Either ABN or WPN must be present",
                    Location = "/tns:TFND/tns:RP/tns:AustralianBusinessNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000186"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND118", Value = GetValueOrEmpty(report.TFND118) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND261", Value = GetValueOrEmpty(report.TFND261) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000186
    
            #region VR.ATO.TFND.000189
    
            /*  VR.ATO.TFND.000189
            The Payer Australian Business Number (ABN) has failed the ABN algorithm check.
    
            Legacy Rule Format:
            FailsABNAlgorithm(^TFND118)

            Technical Business Rule Format:
            FailsABNAlgorithm(^TFND118)
    
            Data Elements:
    
            ^TFND118 = TFND:RP:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = FailsABNAlgorithm(report.TFND118);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434223", Severity = ProcessMessageSeverity.Error,
                    Description = @"ABN is invalid.",
                    LongDescription = @"Australian Business Number (ABN) has failed the algorithm check.",
                    Location = "/tns:TFND/tns:RP/tns:AustralianBusinessNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000189"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND118", Value = GetValueOrEmpty(report.TFND118) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000189
    
            #region VR.ATO.TFND.000192
    
            /*  VR.ATO.TFND.000192
            The Payer Withholding Payer Number(WPN) has failed the WPN algorithm check.
    
            Legacy Rule Format:
            FailsWPNAlgorithm(^TFND261)

            Technical Business Rule Format:
            FailsWPNAlgorithm(^TFND261)
    
            Data Elements:
    
            ^TFND261 = TFND:RP:Identifiers.WithholdingPayerNumber.Identifier
            */
            assertion = FailsTFNAlgorithm(report.TFND261);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200012", Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid WPN",
                    Location = "/tns:TFND/tns:RP/tns:WithholdingPayerNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000192"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND261", Value = GetValueOrEmpty(report.TFND261) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000192
    
            #region VR.ATO.TFND.000195
    
            /*  VR.ATO.TFND.000195
            If provided, Payer contact e-mail address must be in a valid format.
    
            Legacy Rule Format:
            (^TFND126 <> BLANK AND NOT IsValidEmail(^TFND126))

            Technical Business Rule Format:
            (^TFND126 <> BLANK AND NOT IsValidEmail(^TFND126))
    
            Data Elements:
    
            ^TFND126 = TFND:RP:ElectronicContact:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TFND126) != true &&  !(IsMatch(report.TFND126, @"^\S.*@.+\.\S+$") ));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500029", Severity = ProcessMessageSeverity.Error,
                    Description = @"Email Address contains invalid text",
                    LongDescription = @"A text character must be one of the following: A to Z a to z 0 to 9 ! @ $ % & * ( ) - _ = [ ] ; : ' "" , . ? / or a space character.",
                    Location = "/tns:TFND/tns:RP/tns:ElectronicContact/tns:ElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000195"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND126", Value = GetValueOrEmpty(report.TFND126) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000195
    
            #region VR.ATO.TFND.000204
    
            /*  VR.ATO.TFND.000204
            When the Payer address is overseas, the State / Territory must not be present.
    
            Legacy Rule Format:
            ((^TFND79 <> NULL AND NotContainsSet(^TFND79,'"au"')) AND (^TFND32 <> NULL))

            Technical Business Rule Format:
            ((^TFND79 <> NULL AND NotContainsSet(^TFND79,'"au"')) AND (^TFND32 <> NULL))
    
            Data Elements:
    
            ^TFND32 = TFND:RP:AddressDetailsBusiness:AddressDetails.StateOrTerritory.Code
    
            ^TFND79 = TFND:RP:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TFND79 != null && !(IsMatch(report.TFND79, @"au",RegexOptions.IgnoreCase)) && report.TFND32 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500011", Severity = ProcessMessageSeverity.Error,
                    Description = @"State and postcode must not be present for non-Australian addresses",
                    LongDescription = @"Country name is present with State or postcode in the address field.  Determine whether the address is Australian or overseas and correct accordingly.",
                    Location = "/tns:TFND/tns:RP/tns:AddressDetailsBusiness/tns:StateOrTerritoryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000204"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND79", Value = GetValueOrEmpty(report.TFND79) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND32", Value = GetValueOrEmpty(report.TFND32) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000204
    
            #region VR.ATO.TFND.000207
    
            /*  VR.ATO.TFND.000207
            When the Payer address is overseas, the Postcode must not be present.
    
            Legacy Rule Format:
            ((^TFND79 <> NULL AND NotContainsSet(^TFND79,'"au"')) AND (^TFND33 <> NULL))

            Technical Business Rule Format:
            ((^TFND79 <> NULL AND NotContainsSet(^TFND79,'"au"')) AND (^TFND33 <> NULL))
    
            Data Elements:
    
            ^TFND33 = TFND:RP:AddressDetailsBusiness:AddressDetails.Postcode.Text
    
            ^TFND79 = TFND:RP:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TFND79 != null && !(IsMatch(report.TFND79, @"au",RegexOptions.IgnoreCase)) && report.TFND33 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000480", Severity = ProcessMessageSeverity.Error,
                    Description = @"Country code must be blank or set to Australia",
                    Location = "/tns:TFND/tns:RP/tns:AddressDetailsBusiness/tns:PostcodeT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000207"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND79", Value = GetValueOrEmpty(report.TFND79) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND33", Value = GetValueOrEmpty(report.TFND33) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000207
    
            #region VR.ATO.TFND.000210
    
            /*  VR.ATO.TFND.000210
            When the Payer address is within Australia (Country Code is ‘au’ or not supplied), the Postcode must be provided.
    
            Legacy Rule Format:
            ((ContainsSet(^TFND79,'"au"') OR ^TFND79 = NULL) AND ^TFND33 = NULL)

            Technical Business Rule Format:
            ((ContainsSet(^TFND79,'"au"') OR ^TFND79 = NULL) AND ^TFND33 = NULL)
    
            Data Elements:
    
            ^TFND33 = TFND:RP:AddressDetailsBusiness:AddressDetails.Postcode.Text
    
            ^TFND79 = TFND:RP:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = ((IsMatch(report.TFND79, @"au",RegexOptions.IgnoreCase) || report.TFND79 == null) && report.TFND33 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000009", Severity = ProcessMessageSeverity.Error,
                    Description = @"Postcode is required for an Australian address",
                    Location = "/tns:TFND/tns:RP/tns:AddressDetailsBusiness/tns:PostcodeT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000210"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND79", Value = GetValueOrEmpty(report.TFND79) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND33", Value = GetValueOrEmpty(report.TFND33) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000210
    
            #region VR.ATO.TFND.000213
    
            /*  VR.ATO.TFND.000213
            Payer business address country code must consist of a valid Country Code
    
            Legacy Rule Format:
            (^TFND79 <> NULL AND (NotInDomain(^TFND79, 'COUNTRY CODES')))

            Technical Business Rule Format:
            (^TFND79 <> NULL AND (NotInDomain(^TFND79, 'COUNTRY CODES')))
    
            Data Elements:
    
            ^TFND79 = TFND:RP:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TFND79 != null && !(IsMatch(report.TFND79, @"^(ad|ae|af|ag|ai|al|am|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bl|bm|bn|bo|bq|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cu|cv|cw|cx|cy|cz|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mf|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|ss|st|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tr|tt|tv|tw|tz|ua|ug|um|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|za|zm|zw)$",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000213", Severity = ProcessMessageSeverity.Error,
                    Description = @"Payer business address country code is invalid.",
                    Location = "/tns:TFND/tns:RP/tns:AddressDetailsBusiness/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000213"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND79", Value = GetValueOrEmpty(report.TFND79) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000213
    
            #region VR.ATO.TFND.000216
    
            /*  VR.ATO.TFND.000216
            The Payer date of declaration signed cannot be later than today.
    
            Legacy Rule Format:
            ^TFND110 > Today()

            Technical Business Rule Format:
            ^TFND110 > Today()
    
            Data Elements:
    
            ^TFND110 = TFND:RP:Declaration:Declaration.Signature.Date
            */
            assertion = (report.TFND110.GetValueOrDefault() > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000216", Severity = ProcessMessageSeverity.Error,
                    Description = @"The date entered for Payer date of declaration signed must not be later than today's date.",
                    Location = "/tns:TFND/tns:RP/tns:Declaration/tns:SignatureD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000216"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND110", Value = GetValueOrEmptyDate(report.TFND110) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000216
    
            #region VR.ATO.TFND.000219
    
            /*  VR.ATO.TFND.000219
            Payer declaration accepted indicator must be 'TRUE'.
    
            Legacy Rule Format:
            ((^TFND257 <> NULL) AND (^TFND112 <> TRUE))

            Technical Business Rule Format:
            ((^TFND257 <> NULL) AND (^TFND112 <> TRUE))
    
            Data Elements:
    
            ^TFND112 = TFND:RP:Declaration:Declaration.StatementAccepted.Indicator
    
            ^TFND257 = TFND:RP:Declaration
            */
            assertion = (report.RP_DeclarationCollectionExists != false && report.TFND112 != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000219", Severity = ProcessMessageSeverity.Error,
                    Description = @"Payer declaration accepted indicator must be 'TRUE'.",
                    Location = "/tns:TFND/tns:RP/tns:Declaration/tns:StatementAcceptedI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000219"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND112", Value = GetValueOrEmpty(report.TFND112) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000219
    
            #region VR.ATO.TFND.000222
    
            /*  VR.ATO.TFND.000222
            The Payee Tax File Number (TFN) has failed the algorithm check.
    
            Legacy Rule Format:
            ((^TFND38 <> 987654321 ) AND FailsTFNAlgorithm(^TFND38))

            Technical Business Rule Format:
            ((^TFND38 <> 987654321 ) AND FailsTFNAlgorithm(^TFND38))
    
            Data Elements:
    
            ^TFND38 = TFND:Payee:Identifiers:Identifiers.TaxFileNumber.Identifier
            */
            assertion = (AsNumeric(report.TFND38) != 987654321 && FailsTFNAlgorithm(report.TFND38));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.402043", Severity = ProcessMessageSeverity.Error,
                    Description = @"TFN must be a valid TFN",
                    Location = "/tns:TFND/tns:Payee/tns:Identifiers/tns:TaxFileNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000222"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND38", Value = GetValueOrEmpty(report.TFND38) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000222
    
            #region VR.ATO.TFND.000228
    
            /*  VR.ATO.TFND.000228
            Either the Payee TFN or TFN Exemption Code must be provided.
    
            Legacy Rule Format:
            (^TFND38 <> NULL AND ^TFND89 <> NULL) OR (^TFND38 = NULL AND ^TFND89 = NULL)

            Technical Business Rule Format:
            (^TFND38 <> NULL AND ^TFND89 <> NULL) OR (^TFND38 = NULL AND ^TFND89 = NULL)
    
            Data Elements:
    
            ^TFND89 = TFND:Payee:Identifiers:IdentificationExemptionDetails.TFNExemptionType.Code
    
            ^TFND38 = TFND:Payee:Identifiers:Identifiers.TaxFileNumber.Identifier
            */
            assertion = (report.TFND38 != null && report.TFND89 != null || report.TFND38 == null && report.TFND89 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000007", Severity = ProcessMessageSeverity.Error,
                    Description = @"Either Payee Tax File Number or TFN Exemption Code must be present",
                    Location = "/tns:TFND/tns:Payee/tns:Identifiers/tns:IdentificationExemptionDetailsTFNExemptionTypeC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000228"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND38", Value = GetValueOrEmpty(report.TFND38) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND89", Value = GetValueOrEmpty(report.TFND89) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000228
    
            #region VR.ATO.TFND.000231
    
            /*  VR.ATO.TFND.000231
            When Payee day of Birth, Payee month of Birth and Payee year of Birth are all supplied, it must be a valid date.
    
            Legacy Rule Format:
            (^TFND119 <> NULL AND ^TFND220 <> NULL AND ^TFND221 <> NULL AND (IsDate(ConvertToDate(^TFND119, ^TFND220, ^TFND221)) = 'FALSE' ))

            Technical Business Rule Format:
            (^TFND119 <> NULL AND ^TFND220 <> NULL AND ^TFND221 <> NULL AND (IsDate(ConvertToDate(^TFND119, ^TFND220, ^TFND221)) = 'FALSE' ))
    
            Data Elements:
    
            ^TFND119 = TFND:Payee:PersonDemographicDetails:PersonDemographicDetails.Birth.DayofMonth
    
            ^TFND220 = TFND:Payee:PersonDemographicDetails:PersonDemographicDetails.Birth.Month
    
            ^TFND221 = TFND:Payee:PersonDemographicDetails:PersonDemographicDetails.Birth.Year
            */
            assertion = (report.TFND119 != null && report.TFND220 != null && report.TFND221 != null && IsDate(ConvertToDate(report.TFND119.GetValueOrDefault(), report.TFND220.GetValueOrDefault(), report.TFND221.GetValueOrDefault())) == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000231", Severity = ProcessMessageSeverity.Error,
                    Description = @"Payee date of birth is invalid.",
                    Location = "/tns:TFND/tns:Payee/tns:PersonDemographicDetails/tns:BirthDm",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000231"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND119", Value = GetValueOrEmpty(report.TFND119) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND220", Value = GetValueOrEmpty(report.TFND220) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND221", Value = GetValueOrEmpty(report.TFND221) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000231
    
            #region VR.ATO.TFND.000233
    
            /*  VR.ATO.TFND.000233
            When the Payee address is overseas, the State / Territory must not be present.
    
            Legacy Rule Format:
            ((^TFND83 <> NULL AND NotContainsSet(^TFND83,'"au"')) AND (^TFND49 <> NULL))

            Technical Business Rule Format:
            ((^TFND83 <> NULL AND NotContainsSet(^TFND83,'"au"')) AND (^TFND49 <> NULL))
    
            Data Elements:
    
            ^TFND49 = TFND:Payee:AddressDetailsResidential:AddressDetails.StateOrTerritory.Code
    
            ^TFND83 = TFND:Payee:AddressDetailsResidential:AddressDetails.Country.Code
            */
            assertion = (report.TFND83 != null && !(IsMatch(report.TFND83, @"au",RegexOptions.IgnoreCase)) && report.TFND49 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500011", Severity = ProcessMessageSeverity.Error,
                    Description = @"State and postcode must not be present for non-Australian addresses",
                    LongDescription = @"Country name is present with State or postcode in the address field.  Determine whether the address is Australian or overseas and correct accordingly.",
                    Location = "/tns:TFND/tns:Payee/tns:AddressDetailsResidential/tns:StateOrTerritoryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000233"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND83", Value = GetValueOrEmpty(report.TFND83) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND49", Value = GetValueOrEmpty(report.TFND49) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000233
    
            #region VR.ATO.TFND.000234
    
            /*  VR.ATO.TFND.000234
            Payee date of birth cannot be a future date.
    
            Legacy Rule Format:
            ((ConvertToDate(^TFND119, ^TFND220, ^TFND221) <> NULL) AND (ConvertToDate(^TFND119, ^TFND220, ^TFND221) > Today()))

            Technical Business Rule Format:
            ((ConvertToDate(^TFND119, ^TFND220, ^TFND221) <> NULL) AND (ConvertToDate(^TFND119, ^TFND220, ^TFND221) > Today()))
    
            Data Elements:
    
            ^TFND221 = TFND:Payee:PersonDemographicDetails:PersonDemographicDetails.Birth.Year
    
            ^TFND119 = TFND:Payee:PersonDemographicDetails:PersonDemographicDetails.Birth.DayofMonth
    
            ^TFND220 = TFND:Payee:PersonDemographicDetails:PersonDemographicDetails.Birth.Month
            */
            assertion = (ConvertToDate(report.TFND119.GetValueOrDefault(), report.TFND220.GetValueOrDefault(), report.TFND221.GetValueOrDefault()) != null && ConvertToDate(report.TFND119.GetValueOrDefault(), report.TFND220.GetValueOrDefault(), report.TFND221.GetValueOrDefault()) > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000234", Severity = ProcessMessageSeverity.Error,
                    Description = @"Payee date of birth can not be a future date.",
                    Location = "/tns:TFND/tns:Payee/tns:PersonDemographicDetails/tns:BirthY",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000234"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND119", Value = GetValueOrEmpty(report.TFND119) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND220", Value = GetValueOrEmpty(report.TFND220) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND221", Value = GetValueOrEmpty(report.TFND221) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000234
    
            #region VR.ATO.TFND.000236
    
            /*  VR.ATO.TFND.000236
            When the Payee address is overseas, the Postcode must not be present.
    
            Legacy Rule Format:
            ((^TFND83 <> NULL AND NotContainsSet(^TFND83,'"au"')) AND (^TFND50 <> NULL))

            Technical Business Rule Format:
            ((^TFND83 <> NULL AND NotContainsSet(^TFND83,'"au"')) AND (^TFND50 <> NULL))
    
            Data Elements:
    
            ^TFND50 = TFND:Payee:AddressDetailsResidential:AddressDetails.Postcode.Text
    
            ^TFND83 = TFND:Payee:AddressDetailsResidential:AddressDetails.Country.Code
            */
            assertion = (report.TFND83 != null && !(IsMatch(report.TFND83, @"au",RegexOptions.IgnoreCase)) && report.TFND50 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000480", Severity = ProcessMessageSeverity.Error,
                    Description = @"Country code must be blank or set to Australia",
                    Location = "/tns:TFND/tns:Payee/tns:AddressDetailsResidential/tns:PostcodeT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000236"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND83", Value = GetValueOrEmpty(report.TFND83) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND50", Value = GetValueOrEmpty(report.TFND50) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000236
    
            #region VR.ATO.TFND.000237
    
            /*  VR.ATO.TFND.000237
            When the Payee address is within Australia (Country Code is ‘au’ or not supplied), the State / Territory must be provided.
    
            Legacy Rule Format:
            ((^TFND83 = NULL OR ContainsSet(^TFND83,'"au"')) AND (^TFND49 = NULL))

            Technical Business Rule Format:
            ((^TFND83 = NULL OR ContainsSet(^TFND83,'"au"')) AND (^TFND49 = NULL))
    
            Data Elements:
    
            ^TFND49 = TFND:Payee:AddressDetailsResidential:AddressDetails.StateOrTerritory.Code
    
            ^TFND83 = TFND:Payee:AddressDetailsResidential:AddressDetails.Country.Code
            */
            assertion = ((report.TFND83 == null || IsMatch(report.TFND83, @"au",RegexOptions.IgnoreCase)) && report.TFND49 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000112", Severity = ProcessMessageSeverity.Error,
                    Description = @"For an Australian address, the state or territory must be provided",
                    Location = "/tns:TFND/tns:Payee/tns:AddressDetailsResidential/tns:StateOrTerritoryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000237"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND83", Value = GetValueOrEmpty(report.TFND83) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND49", Value = GetValueOrEmpty(report.TFND49) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000237
    
            #region VR.ATO.TFND.000239
    
            /*  VR.ATO.TFND.000239
            When the Payee address is within Australia (Country Code is ‘au’ or not supplied), the Postcode must be provided.
    
            Legacy Rule Format:
            ((ContainsSet(^TFND83,'"au"') OR ^TFND83 = NULL) AND ^TFND50 = NULL)

            Technical Business Rule Format:
            ((ContainsSet(^TFND83,'"au"') OR ^TFND83 = NULL) AND ^TFND50 = NULL)
    
            Data Elements:
    
            ^TFND50 = TFND:Payee:AddressDetailsResidential:AddressDetails.Postcode.Text
    
            ^TFND83 = TFND:Payee:AddressDetailsResidential:AddressDetails.Country.Code
            */
            assertion = ((IsMatch(report.TFND83, @"au",RegexOptions.IgnoreCase) || report.TFND83 == null) && report.TFND50 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000009", Severity = ProcessMessageSeverity.Error,
                    Description = @"Postcode is required for an Australian address",
                    Location = "/tns:TFND/tns:Payee/tns:AddressDetailsResidential/tns:PostcodeT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000239"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND83", Value = GetValueOrEmpty(report.TFND83) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND50", Value = GetValueOrEmpty(report.TFND50) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000239
    
            #region VR.ATO.TFND.000242
    
            /*  VR.ATO.TFND.000242
            Payee residential address country code must consist of a valid Country Code
    
            Legacy Rule Format:
            (^TFND83 <> NULL AND (NotInDomain(^TFND83, 'COUNTRY CODES')))

            Technical Business Rule Format:
            (^TFND83 <> NULL AND (NotInDomain(^TFND83, 'COUNTRY CODES')))
    
            Data Elements:
    
            ^TFND83 = TFND:Payee:AddressDetailsResidential:AddressDetails.Country.Code
            */
            assertion = (report.TFND83 != null && !(IsMatch(report.TFND83, @"^(ad|ae|af|ag|ai|al|am|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bl|bm|bn|bo|bq|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cu|cv|cw|cx|cy|cz|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mf|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|ss|st|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tr|tt|tv|tw|tz|ua|ug|um|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|za|zm|zw)$",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000242", Severity = ProcessMessageSeverity.Error,
                    Description = @"Payee residential address country code is invalid.",
                    Location = "/tns:TFND/tns:Payee/tns:AddressDetailsResidential/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000242"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND83", Value = GetValueOrEmpty(report.TFND83) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000242
    
            #region VR.ATO.TFND.000244
    
            /*  VR.ATO.TFND.000244
            The Payee date declaration signed cannot be later than today.
    
            Legacy Rule Format:
            ^TFND62 > Today()

            Technical Business Rule Format:
            ^TFND62 > Today()
    
            Data Elements:
    
            ^TFND62 = TFND:Payee:DeclarationHardCopy:Declaration.Signature.Date
            */
            assertion = (report.TFND62.GetValueOrDefault() > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000244", Severity = ProcessMessageSeverity.Error,
                    Description = @"The date entered for Payee date declaration signed must not be later than today's date.",
                    Location = "/tns:TFND/tns:Payee/tns:DeclarationHardCopy/tns:SignatureD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000244"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND62", Value = GetValueOrEmptyDate(report.TFND62) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000244
    
            #region VR.ATO.TFND.000250
    
            /*  VR.ATO.TFND.000250
            When Supplier is provided then Supplier Declaration must be present.
    
            Legacy Rule Format:
            ((^TFND200 <> NULL) AND (^TFND260 = NULL))

            Technical Business Rule Format:
            ((^TFND200 <> NULL) AND (^TFND260 = NULL))
    
            Data Elements:
    
            ^TFND200 = TFND:INT
    
            ^TFND260 = TFND:INT:Declaration
            */
            assertion = (report.IntCollectionExists != false && report.InT_DeclarationCollectionExists == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000250", Severity = ProcessMessageSeverity.Error,
                    Description = @"When Supplier is provided then Supplier Declaration must be present.",
                    Location = "/tns:TFND/tns:INT/tns:Declaration/tns:StatementAcceptedI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000250"} },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000250
    
            #region VR.ATO.TFND.000253
    
            /*  VR.ATO.TFND.000253
            When Supplier is not provided then Payer Declaration must be present.
    
            Legacy Rule Format:
            ((^TFND200 = NULL) AND (^TFND257 = NULL))

            Technical Business Rule Format:
            ((^TFND200 = NULL) AND (^TFND257 = NULL))
    
            Data Elements:
    
            ^TFND200 = TFND:INT
    
            ^TFND257 = TFND:RP:Declaration
            */
            assertion = (report.IntCollectionExists == false && report.RP_DeclarationCollectionExists == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000253", Severity = ProcessMessageSeverity.Error,
                    Description = @"When Supplier is not provided then Payer Declaration must be present.",
                    Location = "/tns:TFND/tns:RP/tns:Declaration/tns:SignatureD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000253"} },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000253
    
            #region VR.ATO.TFND.000260
    
            /*  VR.ATO.TFND.000260
            When Supplier Business Address Details is provided, and the address is within Australia (Country Code is ‘au’ or not supplied),  the State / Territory must be provided.
    
            Legacy Rule Format:
            ^TFND206 <> NULL AND (^TFND75 = NULL OR ContainsSet(^TFND75,'"au"')) AND ^TFND19 = NULL

            Technical Business Rule Format:
            ^TFND206 <> NULL AND (^TFND75 = NULL OR ContainsSet(^TFND75,'"au"')) AND ^TFND19 = NULL
    
            Data Elements:
    
            ^TFND19 = TFND:INT:AddressDetailsBusiness:AddressDetails.StateOrTerritory.Code
    
            ^TFND75 = TFND:INT:AddressDetailsBusiness:AddressDetails.Country.Code
    
            ^TFND206 = TFND:INT:AddressDetailsBusiness
            */
            assertion = (report.InT_AddressDetailsBusinessCollectionExists != false && (report.TFND75 == null || IsMatch(report.TFND75, @"au",RegexOptions.IgnoreCase)) && report.TFND19 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TFND.000112", Severity = ProcessMessageSeverity.Error,
                    Description = @"For an Australian address, the state or territory must be provided",
                    Location = "/tns:TFND/tns:INT/tns:AddressDetailsBusiness/tns:StateOrTerritoryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000260"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND75", Value = GetValueOrEmpty(report.TFND75) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND19", Value = GetValueOrEmpty(report.TFND19) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000260
    
            #region VR.ATO.TFND.000270
    
            /*  VR.ATO.TFND.000270
            If provided, Payee Primary e-mail address must be in a valid format.
    
            Legacy Rule Format:
            (^TFND271 <> BLANK AND NOT IsValidEmail(^TFND271))

            Technical Business Rule Format:
            (^TFND271 <> BLANK AND NOT IsValidEmail(^TFND271))
    
            Data Elements:
    
            ^TFND271 = TFND:Payee:EmailAddress:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TFND271) != true &&  !(IsMatch(report.TFND271, @"^\S.*@.+\.\S+$") ));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500029", Severity = ProcessMessageSeverity.Error,
                    Description = @"Email Address contains invalid text",
                    LongDescription = @"A text character must be one of the following: A to Z a to z 0 to 9 ! @ $ % & * ( ) - _ = [ ] ; : ' "" , . ? / or a space character.",
                    Location = "/tns:TFND/tns:Payee/tns:EmailAddressElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TFND.000270"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TFND271", Value = report.TFND271 });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TFND.000270

            foreach (ProcessMessageDocument currentProcessMessage in response)
            {
                if (currentProcessMessage.Parameters != null)
                {
                    foreach (ProcessMessageParameter currentParameter in currentProcessMessage.Parameters)
                    {
                        if (string.IsNullOrEmpty(currentParameter.Name))
                        {
                            currentParameter.Name = _emptyParameterValue;
                        }

                        if (currentParameter.Name.Length > _maxParameterNameLength)
                        {
                            currentParameter.Name = currentParameter.Name.Substring(0, _maxParameterNameLength - 1);
                        }

                        if (string.IsNullOrEmpty(currentParameter.Value))
                        {
                            currentParameter.Value = _emptyParameterValue;
                        }

                        if (currentParameter.Value.Length > _maxParameterValueLength)
                        {
                            currentParameter.Value = currentParameter.Value.Substring(0, _maxParameterValueLength - 1);
                        }
                    }
                }
            }

            return response;
        }


    }
} 
